"""

Part of the tagit module.
A copy of the license is provided with the project.
Author: Matthias Baumgartner, 2016
"""
# imports

# inner-module imports
from proxy import Proxy
from tagit.bindings import Binding

# exports
__all__ = ('NextPage', 'PreviousPage', 'ScrollUp', 'ScrollDown', 'FirstImage', 'LastImage', 'RotateLeft', 'RotateRight', 'Delete')


## CODE ##

###########################################################################
#                               NAVIGATION                                #
###########################################################################

class NextPage(Proxy):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'page_next', Binding.simple(Binding.PGDOWN))):
            self.apply()
            return True
        return False

    def apply(self):
        """Page Down
        Cursor: No change
        View: Jump by N*M images
        """
        self = self._buddy # Proxy hook
        page_size = self.page_size()
        if self.offset + page_size < self.n_results:
            self.offset += page_size
            self.parent.dispatch('on_page_change', self, self.offset)
            self.redraw()


class PreviousPage(Proxy):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'page_prev', Binding.simple(Binding.PGUP))):
            self.apply()
            return True
        return False

    def apply(self):
        """Page Up
        Cursor: No change
        View: Jump by N*M images
        """
        self = self._buddy # Proxy hook
        self.offset -= self.cols * self.rows
        self.offset = max(self.offset, 0)
        self.parent.dispatch('on_page_change', self, self.offset)
        self.redraw()


class ScrollUp(Proxy):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'scroll_up', Binding.simple(107, None, Binding.ALL))): # k
            self.apply()
            return True
        return False

    def apply(self):
        """Scroll Up
        Cursor: No change
        View: Jump one row up
        """
        self = self._buddy # Proxy hook
        self.offset -= self.cols
        self.offset = max(self.offset, 0)
        self.parent.dispatch('on_page_change', self, self.offset)
        self.redraw()


class ScrollDown(Proxy):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'scroll_down', Binding.simple(106, None, Binding.ALL))): # j
            self.apply()
            return True
        return False

    def apply(self):
        """Scroll Down
        Cursor: No change
        View: Jump one row down
        """
        self = self._buddy # Proxy hook
        step = self.cols
        page_size = self.page_size()
        if self.offset + step + page_size < self.n_results + self.cols:
            # Cannot scroll below last full page
            # Remove the *page_size* and *self.cols* terms to weaken this property.
            self.offset += step
            self.parent.dispatch('on_page_change', self, self.offset)
        self.redraw()


class FirstImage(Proxy):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'go_first', Binding.simple(Binding.HOME))):
            self.apply()
            return True
        return False

    def apply(self):
        """Home
        Cursor: Jump to first image
        View: Jump to first page
        """
        self = self._buddy # Proxy hook
        self.offset = 0
        self.parent.dispatch('on_page_change', self, self.offset)
        self.set_cursor_first()
        self.redraw()


class LastImage(Proxy):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'go_last', Binding.simple(Binding.END))):
            self.apply()
            return True
        return False

    def apply(self):
        """End
        Cursor: Jump to last image
        View: Jump to last page
        """
        self = self._buddy # Proxy hook
        page_size = self.page_size()
        row_size = self.cols

        #self.offset = self.n_results - 1 # Pick last image

        # ceil(n_results / page_size) -> number of (full) pages
        # page size * number of (full) pages -> number of images to fill up all pages
        #self.offset = int(page_size * ceil(1.0 * self.n_results / page_size) - page_size) # Pick last page

        # n_results % row_size -> images on last row
        # row_size - (images on last row) % row_size -> missing images on last row
        self.offset = self.n_results + ((row_size - (self.n_results % row_size)) % row_size) - page_size # Pick last row

        self.parent.dispatch('on_page_change', self, self.offset)
        self.set_cursor_last()
        self.redraw()


class Rotate(Proxy):
    def rotate(self, spin):
        """Rotate selected images clockwise (spin=-1) or counter-clockwise (spin=1).
        """
        self = self._buddy # Proxy hook
        if len(self.selection) > 0:
            w0 = [1, 8, 3, 6]
            w1 = [2, 7, 4, 5]
            f0 = lambda c: w0[(w0.index(c) + spin) % 4]
            f1 = lambda c: w1[(w1.index(c) + spin) % 4]

            for img in self.selection:
                curr = self.model.attributes.orientation(img)
                curr = (curr is not None and curr > 0) and curr or 1
                ori = curr in w0 and f0(curr) or f1(curr)
                self.model.attributes.set_orientation(img, ori)


###########################################################################
#                                 ACTIONS                                 #
###########################################################################

class RotateLeft(Rotate):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'rotate_left', Binding.simple(Binding.DOT, Binding.SHIFT))): # >
            self.apply()
            return True
        return False

    def apply(self):
        self.rotate(1)
        self._buddy.redraw()


class RotateRight(Rotate):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'rotate_right', Binding.simple(Binding.COMMA, Binding.SHIFT))): # <
            self.apply()
            return True
        return False

    def apply(self):
        self.rotate(-1)
        self._buddy.redraw()


class Delete(Proxy):
    def on_keyboard(self, wx, evt):
        if Binding.check(evt, self._buddy.settings.trace('bindings', 'browser', 'delete', Binding.simple(Binding.DEL, Binding.CTRL, Binding.REST))): # CTRL + DEL
            self._buddy.request_delete()
            return True

    def apply(self):
        self = self._buddy # Proxy hook
        self.request_delete()

## EOF ##
