"""Proxy.

A proxy enables a controller to generalize some of its functionality. An action implements a
specific aspect of the controller.

The action combines the input methods (keyboard, mouse) and the effect in the controller. As
unit, it can be easily utilized from the UI. Furthermore, the same action can be referred to
by multiple UI elements.

Proxies link closely to the controller and allow flexible extension of a controller without
overloading it.

Proxy classes assume perfect knowledge of their buddy (controller). UI elements work with the
usual abstraction between controller and view. Like so, the two can be exchanged.

The Proxy builds upon on the controller's event handling and the controller hierarchy.

The following use case describes intention and layout of a proxy:

    controller:
        browser.py
            rbinds to proxy
            implements proxy factory (get_proxy)

        proxy:
            A series of small classes, each implementing the Proxy defined in this file.
            Apply functions use the proxy hook:
                self = self._buddy
            to ease the notation

    view:
        status.kv
            includes a set of actions

        actions:
            buttons or ui elements can be defined in the kivy file.
            The classes must mixin the proxy (proxy_mixin) and set the
            variables RCPT and PROXY (as required by the mixin).


    * When the application is built, browser and actions are created and placed in the UI.
      The browser rbinds to the proxy signal
    * Before startup, actions from the UI initialize the respective proxy by requesting
      it from the browser.
    * When a button is clicked, the apply function of the proxy is executed
    * When a key is pressed, it is routed to the proxy via browser


Part of the tagit module.
A copy of the license is provided with the project.
Author: Matthias Baumgartner, 2016
"""
# imports

# inner-module imports

# exports
__all__ = ('Proxy', 'recursive_proxy')


## CODE ##

class Proxy(object):
    """Proxy interface."""
    def __init__(self, buddy):
        self._buddy = buddy

    def apply(self):
        # Use the following line!
        #    self = self._buddy # Proxy hook
        raise NotImplementedError()

    def on_keyboard(self, *args, **kwargs):
        raise NotImplementedError()

def recursive_proxy(ctrl, *args, **kwargs):
    """Search for proxy starting at *ctrl*, working its way up the
    controller hierarchy. Return None if proxy not found.
    """
    while ctrl is not None:
        proxy = ctrl.request('proxy', *args, **kwargs)
        if len(proxy) == 0:
            ctrl = ctrl.parent # search up
        else:
            return proxy[0]

    return None

## EOF ##
