"""Suggestions controller.

The class has two modes:
* Add tags to a filter query
* Add tags to selected images

and the following functionality:
* Suggest tags for current view
* Blacklist tags from suggestion until next insert

The UI is a list of tags

UI:
    * swipe right: blacklist tag
    * tap: add tag to {selected images, filter list}

    blacklist:
    * don't show tag when suggested
    * until next adding

    add tag:
    * add tag to images
    * rebuild suggestions > new items

    add filter:
    * add tag to filter list
    * rebuild suggestions > present items

    rebuild suggestions:
    * somehow get suggestions
    * remove blacklisted
    * empty blacklist
    * top N tags

TODO: Put the top ranked tag in the center (instead the top) !!

Part of the tagit module.
A copy of the license is provided with the project.
Author: Matthias Baumgartner, 2017

"""
# imports
from os.path import expanduser

# inner-module imports
from ..basics import difference
from controller import DataController
from ..bindings import Binding
from ..token import token_factory
from filter import Frame
from tagit.suggestion import suggest_rank, suggest_new
from tagit.datalog import DataLog

Token_Tag = token_factory('tag')

# exports
__all__ = ('CSuggestion', )

## code ##

class CSuggestion(DataController):
    MODE_SUGGEST = 'suggest'
    MODE_FILTER  = 'filter'

    def __init__(self, widget, model, settings, parent=None, **kwargs):
        super(CSuggestion, self).__init__(widget, model, settings, parent, **kwargs)
        self.log = DataLog(expanduser(self.settings.trace('session', 'searchlog', '~/.taglog.db')))
        self.mode = CSuggestion.MODE_FILTER
        self.active = []
        self.selection = []
        self.blacklist = {
            CSuggestion.MODE_FILTER: [],
            CSuggestion.MODE_SUGGEST: []
        }
        self.cache = {
                CSuggestion.MODE_SUGGEST: self.suggest_tag(),
                CSuggestion.MODE_FILTER: self.suggest_filter()
        }

        # Event bindings
        self.parent.bind(on_selection=self._upd_selection)

        # Keyboard bindings
        self.get_root().bind(on_keyboard=self.on_keyboard)

    def __del__(self):
        # Release bindings
        self.get_root().unbind(on_keyboard=self.on_keyboard)

    def _upd_selection(self, browser, selection):
        self.selected = selection


    ###########################################################################
    #                                 GETTERS                                 #
    ###########################################################################

    def num_tokens(self):
        return len(self.active)

    def apply(self):
        """Run the filter.
        """
        # query the filter from the database
        query_result = self.model.query(self.active)
        self.log.log_search( # FIXME: Need to serialize tokens (i.e. require type as well) so that they can be restored!

            self.model.get_instance_infos(),
            map(lambda t: t.edit(), self.active),
            [] #map(lambda t: t.edit(), self.inactive)
            )

    # Propagate the result via the parent
        self.parent.dispatch('on_results_change', self, query_result, Frame())
        return False # Don't prevent other listeners to run.

    ###########################################################################
    #                               KEYBINDINGS                               #
    ###########################################################################

    def on_keyboard(self, wx, evt):
        """Handle filter keyboard events.
        """
        if Binding.check(evt, self.settings.trace('bindings', 'filter', 'redraw     ', Binding.simple(286))): # F5
            self.apply()
            return True
        elif Binding.check(evt, self.settings.trace('bindings', 'filter', 'go_back    ', Binding.multi((Binding.BACKSPACE, Binding.CTRL, Binding.REST), (Binding.LEFT, Binding.ALT, Binding.REST)))):
            self.go_back(1)
            return True

        return False

    ###########################################################################
    #                               MOUSE EVENTS                              #
    ###########################################################################

    def go_back(self, nSteps=1):
        """Remove the last *nSteps* tokens.
        """
        self.active = self.active[:-nSteps]
        self.apply()
        sugg = self.suggest_filter()
        self.cache[self.mode] = sugg
        self.widget.redraw(sugg)

    def add_item(self, item):
        if self.mode == CSuggestion.MODE_SUGGEST:
            sugg = self.add_tag(item)
        elif self.mode == CSuggestion.MODE_FILTER:
            sugg = self.add_filter(item)

        self.cache[self.mode] = sugg
        self.widget.redraw(sugg)

    def add_tag(self, tag):
        """
        """
        # add tag to images
        self.model.tags.set(self.selection, tag)
        return self.suggest_tag()

    def suggest_tag(self):
        # assemble suggestions
        sugg = suggest_new(self.log, self.model, self.model.query(self.active))
        # exclude blacklisted
        sugg = difference(sugg, self.blacklist[CSuggestion.MODE_SUGGEST])
        self.blacklist[CSuggestion.MODE_SUGGEST] = []
        # exclude tags present in all images (excludes active ones too)
        #imgs = self.model.query(self.active)
        #hist = self.model.tags.histogram(include=imgs)
        #omni = [k for k, v in hist.iteritems() if v >= len(imgs)]
        #sugg = difference(sugg, omni)

        return sugg

    def add_filter(self, item):
        tag = Token_Tag().from_string(item)
        self.active.append(tag)
        self.apply()
        return self.suggest_filter()

    def suggest_filter(self):
        imgs = self.model.query(self.active)
        sugg = self.model.tags.get(imgs)
        # exclude tags present in all images
        hist = self.model.tags.histogram(include=imgs)
        omni = [k for k, v in hist.iteritems() if v >= len(imgs)]
        sugg = difference(sugg, omni)
        # exclude active tags
        filt = map(lambda t: t.get(), self.active)
        sugg = difference(sugg, filt)
        # exclude blacklisted
        sugg = difference(sugg, self.blacklist[CSuggestion.MODE_FILTER])
        self.blacklist[CSuggestion.MODE_FILTER] = []
        # TODO Magic: Order/Rank tags
        sugg.sort(key=suggest_rank(sugg))

        return sugg

    def add_blacklist(self, item):
        self.blacklist[self.mode].append(item)
        #if item in self.cache[self.mode]: self.cache[self.mode].remove(item)
        #self.widget.redraw(self.cache[self.mode])

    def set_mode_filter(self):
        self.mode = CSuggestion.MODE_FILTER
        self.widget.redraw(self.cache[self.mode])

    def set_mode_suggest(self):
        self.mode = CSuggestion.MODE_SUGGEST
        self.widget.redraw(self.cache[self.mode])

## EOF ##
