"""

Part of the tagit module.
A copy of the license is provided with the project.
Author: Matthias Baumgartner, 2017
"""
# imports
import numpy as np
from scipy.sparse import dok_matrix
from tagit.external.memoize import memoized

# exports
__all__ = ('Graph', )
## code ##

list_type = lambda obj: isinstance(obj, list) or isinstance(obj, set) or isinstance(obj, tuple)

class Graph(object):
    def __init__(self, conn):

        self.rows = 1 + conn.execute("SELECT max(id) FROM tag").fetchall()[0][0]
        self.cols = 1 + conn.execute("SELECT max(id) FROM image").fetchall()[0][0]

        self.adj = dok_matrix((self.rows, self.cols))
        for img, tag, t_created in conn.execute("""
            SELECT
                image,
                tag,
                t_created
            FROM
                img_tag
        """): self.adj[int(tag), int(img)] = 1.0 # store t_created instead, then use counts instead of sum.

        self.adjc = self.adj.tocoo().tocsc()
        self.adjr = self.adj.tocoo().tocsr()
        self._conn = conn

    @memoized
    def tag_by_name(self, name):
        r = self._conn.execute("SELECT id FROM tag WHERE tag = ?  ", (name, )).fetchone()
        if r is not None:
            return Tag(int(r[0]), self)
        return Phony(name, self)

    @memoized
    def image_by_name(self, name):
        r = self._conn.execute("SELECT id FROM image WHERE path = ?  ", (name, )).fetchone()
        if r is not None:
            return Image(int(r[0]), self)
        return Phony(name, self)

    def image(self, id_):
        return Image(id_, self)

    def tag(self, id_):
        return Tag(id_, self)

    def n_images(self):
        return self.cols # FIXME: Has to count from matrix (-> only assigned images)

    def n_tags(self):
        return self.rows # FIXME: Has to count from matrix (-> only assigned tags)

class Node(object):
    def __init__(self, id_, parent):
        self._parent = parent
        self._id = id_

    def __str__(self):
        return str(self._id)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self._id)

class Image(Node):
    def __contains__(self, item):
        """Return True if the image is associated with tag *item*.
        If *item* is a list, all its elements must be present.
        """
        if list_type(item):
            return all(i in self for i in item) # recursion
        else:
            return self._parent.adj[item, self._id] > 0

    def __len__(self):
        """Number of tags associated with this image."""
        return self._parent.adjc[:, self._id].nnz

    def __iter__(self):
        """Iterator over tags associated with this image."""
        for itm in self._parent.adjc.getcol(self._id).indices:
            yield self._parent.tag(int(itm))

    @memoized
    def name(self):
        return self._parent._conn.execute("""
            SELECT path
            FROM image
            WHERE id = ?
            """, (self._id, )).fetchall()[0][0]

class Tag(Node):
    def __contains__(self, item):
        """Return True if the tag is associated with image *item*.
        If *item* is a list, all its elements must be present.
        """
        if list_type(item):
            return all(i in self for i in item) # recursion
        else:
            return self._parent.adj[self._id, item] > 0

    def __len__(self):
        """Number of images associated with this tag."""
        return self._parent.adjr[self._id, :].nnz

    def __iter__(self):
        """Iterator over images associated with this tag."""
        for itm in self._parent.adjr.getrow(self._id).indices:
            yield self._parent.image(int(itm))

    @memoized
    def name(self):
        return self._parent._conn.execute("""
            SELECT tag
            FROM tag
            WHERE id = ?
            """, (self._id, )).fetchall()[0][0]

class Phony(Node):
    def __contains__(self, item):
        return False
    def __len__(self):
        return 0
    def __iter__(self):
        return iter([])
    def name(self):
        return self._id
## main ##

if __name__ == '__main__':
    import sqlite3
    import sys

    conn = sqlite3.connect(sys.argv[1])
    g = Graph(conn)
    t = g.tag(143)
    i = list(iter(t))[12]

## EOF ##
